"""
Using the utility class Emva1288, process a descriptor file
print results and plot graphics
"""

import os
from emva1288.process import Emva1288
from emva1288.camera.dataset_generator import DatasetGenerator

# dir_ = '/home/work/1288/datasets/'
# fname = 'EMVA1288_ReferenceSet_001_CCD_12Bit/EMVA1288_Data.txt'
# fname = os.path.join(dir_, fname)

dataset_generator = DatasetGenerator(width=100,
                                     height=50,
                                     bit_depth=8,
                                     dark_current_ref=30)
fname = dataset_generator.descriptor_path

e = Emva1288(fname)
e.results()
e.plot()



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""Transform image data into data
This class takes a dictionary with image data
(product of :class:`emva1288.process.loader.LoadImageData`)
and transforms it into data that can be used for the EMVA1288 computations.
It is important to note, that this is separate from LoadImageData because this
step, produces float values that are not easily transportable (db, json,
etc...) without loosing accuracy.

"""

import numpy as np
import logging


class Data1288(object):
    """Take data from parsed images (descriptor file)
    and expose it as temporal and spatial dicts
    This dicts have the appropriate form for the processing
    """

    def __init__(self,
                 data,
                 pixels=None,
                 loglevel=logging.INFO):
        """Data processing object init method.

        It sets the logging configuration and creates a
        :class:`python:logging.Logger` object using
        :func:`python:logging.getLogger` and sets the loglevel.
        Creates the data dictionaries with the call of
        :meth:`_get_temporal` for temporal data and :meth:`_get_spatial`
        for spatial data.

        Parameters
        ----------
        data : The data descriptor file.
        loglevel : int, optional
                   The loglevel for the Logger object.
        """

        logging.basicConfig()
        self.log = logging.getLogger('Data')
        self.log.setLevel(loglevel)

        self.pixels = pixels or data['width'] * data['height']

        self.data = {}
        self.data['temporal'] = self._get_temporal(data['temporal'])
        self.data['spatial'] = self._get_spatial(data['spatial'])

    def _get_temporal(self, data):
        """Fill the temporal dict, with the stuff that we need.
        Compute the averages and variances from the sums (sum and pvar)

        If there is only one exposure time,
        the arrays in the returned dict will all have the same length as the
        photon count array. For this case, the exposure times and the dark
        value data array elements will all be the same.

        Parameters
        ----------
        data : The data dictionary containing the temporal data sets.

        Returns
        -------
        dict : A dict containing all temporal test data.
               The keys are the following:
               *'texp'*: the array of the
               exposure times used for the test,
               *'u_p'*: the array of
               photon count in a pixel for each exposure time and photon count,
               *'u_y'*: the
               array of the mean digital value for each exposure
               time and photon count,
               *'s2_y'*:
               the array of the digital value variance for each exposure
               time and photon count,
               *'u_ydark'*: the array of the mean digital dark value
               for each exposure time
               and *'s2_ydark'*: the
               array of the digital dark value variance for each exposure time.

        Raises
        ------
        ValueError
            If there is no one 0.0 photon count entry for each exposure time
            and at least one bright point
        """
        temporal = {}

        # List of exposure times
        exposures = np.asarray(sorted(data.keys()))

        # texp is now an array of exposure times
        temporal['texp'] = exposures

        u_p = []
        u_y = []
        s2_y = []
        u_ydark = []
        s2_ydark = []

        for t in exposures:
            # photons is a list of photon counts
            # images for each exposure time
            photons = sorted(data[t].keys())

            if 0.0 not in photons:
                raise ValueError('Every exposure point must have a 0.0 photon')

            if len(photons) < 2:
                raise ValueError('There must be at least one bright photon')

            # get data for dark image
            d = self._get_temporal_data(data[t][0.0])
            u_ydark.append(d['mean'])
            s2_ydark.append(d['var'])

            for p in photons[1:]:
                # For each photon count, get the data
                u_p.append(p)
                d = self._get_temporal_data(data[t][p])
                u_y.append(d['mean'])
                s2_y.append(d['var'])

        # Append all data to temporal dict
        temporal['u_p'] = np.asarray(u_p)
        temporal['u_y'] = np.asarray(u_y)
        temporal['s2_y'] = np.asarray(s2_y)
        temporal['u_ydark'] = np.asarray(u_ydark)
        temporal['s2_ydark'] = np.asarray(s2_ydark)

        # In case we have only one exposure, we need arrays with the
        # same length as the up
        # we just repeat the same value over and over
        if len(exposures) == 1:
            l = len(temporal['u_p'])

            v = temporal['texp'][0]
            temporal['texp'] = np.asarray([v for _i in range(l)])

            v = temporal['u_ydark'][0]
            temporal['u_ydark'] = np.asarray([v for _i in range(l)])

            v = temporal['s2_ydark'][0]
            temporal['s2_ydark'] = np.asarray([v for _i in range(l)])

        return temporal

    def _get_temporal_data(self, d):
        """Convert temporal image data to mean and variance

        The mean is the sum of the pixels of the two images divided by
        (2 * self.pixels)

        The variance is the pseudo variance(integer), divided by
        (4 * self.pixels)

        Parameters
        ----------
        d : dict
            The data dictionary that contains the sum and pvar of the
            pixels of a sum of two consecutive images with
            the same photon count
            and exposure time.

        Returns
        -------
        dict : A data dictionary with the following keys:
               *'mean'*: the mean as described above and
               *'var'*: the variance as described above.
        """
        mean_ = d['sum'] / (2.0 * self.pixels)
        var_ = d['pvar'] / (4.0 * self.pixels)
        return {'mean': mean_, 'var': var_}

    def _get_spatial(self, data):
        """Fill the spatial dictionary.

        The images (sum and pvar) are preserved,
        they are needed for processing.

        Parameters
        ----------
        data : The data dictionary to take data from.

        Returns
        -------
        dict : A dict containing all spatial test data.
               The keys are the following:

               - *'texp'*: the array of exposure times for spatial tests,
               - *'u_p'*: the array of photon count average for
                 each exposure times.
               - *'sum'*: the array of the image sum for each
                 photon count and exposure time,
               - *'pvar'*: the array of the pvar image for each
                 photon count and exposure time,
               - *'L'*: the number of image taken to make the sum and
                 pvar images for each photon count and each exposure time,
               - *'avg'*: the average computed from the sum image for
                 each photon count and each exposure time,
               - *'var'*: the varianve computed from the pvar image for each
                 photon count and each exposure time,
               - *'sum_dark'*: the sum image in the dark for
                 each exposure time,
               - *'pvar_dark'*: the pvar image in the dark
                 for each exposure time,
               - *'L_dark'*: the number of image taken in the dark to
                 compose the sum and pvar images for each exposure time,
               - *'avg_dark'*: the average computed from the dark
                 sum image for each exposure time,
               - *'var_dark'*: the variance computed from the dark
                 pvar image for each exposure time.

        Raises
        ------
        ValueError
            If the there is no exactly one dark and one bright point
        """

        if len(data) != 1:
            raise ValueError('Spatial data must contain only one exposure'
                             ' time')

        # the spatial exposure is the only exposure
        exposure = list(data.keys())[0]

        spatial = {}
        spatial['texp'] = exposure

        photons = sorted(data[exposure].keys())
        if 0.0 not in photons:
            raise ValueError('there must be a 0.0 photon count for spatial')

        if len(photons) != 2:
            raise ValueError('There must be one bright and one dark')

        # get dark spatial data
        d = self._get_spatial_data(data[exposure][0.0], postfix='_dark')
        spatial.update(d)

        # Photon count for bright spatial
        bphoton = photons[1]
        d = self._get_spatial_data(data[exposure][bphoton])
        spatial['u_p'] = bphoton
        spatial.update(d)

        return spatial

    def _get_spatial_data(self, d, postfix=''):
        """Add the mean and variance to the spatial image data.

        The mean is the sum of the images divided by L,
        which is the number of images for the spatial test.

        The variance is the pseudovariance divided by
        (L^2 * (L-1)).

        Parameters
        ----------
        d : dict
            The data dictionary containing the sum and pvar of the images.
        postfix: str, optional
            String to add in the resulting dictionary keys

        Returns
        -------
        dict : A data dictionary processed from the input. The keys (+ postfix)
        are:
               - *'sum'*: the sum image preserved from input,
               - *'pvar'*: the pvar image preserved from input,
               - *'L'*: the number of image summed,
               - *'avg'*: the average of the sum image as described above and
               - *'var'*: the variance as described above computed
                 from the pvar image.
        """

        # This cast is just in case the original images are unint
        sum_ = d['sum'].astype(np.int64)
        pvar_ = d['pvar'].astype(np.int64)

        L = d['L']
        avg_ = sum_ / (1.0 * L)
        var_ = pvar_ / (1.0 * np.square(L) * (L - 1))

        return {'sum' + postfix: sum_,
                'pvar' + postfix: pvar_,
                'L' + postfix: L,
                'avg' + postfix: avg_,
                'var' + postfix: var_}



from emva1288 import process
from emva1288 import report
from emva1288.camera.dataset_generator import DatasetGenerator
import os

# # Load one test to add it as operation point
# dir_ = '/home/work/1288/datasets/'
# fname = 'EMVA1288_ReferenceSet_003_Simulation_12Bit/EMVA1288_Data.txt'
# fname = os.path.join(dir_, fname)

dataset_generator = DatasetGenerator(width=640,
                                     height=480,
                                     K=0.1,
                                     blackoffset=29.4,
                                     bit_depth=12,
                                     steps=50,
                                     exposure_fixed=1000000,
                                     dark_current_ref=30)

fname = dataset_generator.descriptor_path

parser = process.ParseEmvaDescriptorFile(fname)
imgs = process.LoadImageData(parser.images)
dat = process.Data1288(imgs.data)


# Description of the setup
setup = report.info_setup()
setup['Standard version'] = 3.1

# Basic information
basic = report.info_basic()
basic['vendor'] = 'Simulation'
basic['data_type'] = 'Single'
basic['sensor_type'] = 'simulated sensor'
basic['resolution'] = '640x480'
basic['model'] = 'Simulated camera'


# Marketing information
marketing = report.info_marketing()
marketing['watermark'] = 'Example'

# Initialize the report with the marketing data
# Provide a non existent name for the output directory
myreport = report.Report1288('myreport',
                             marketing=marketing,
                             setup=setup,
                             basic=basic)

# Operation point
# bit_depth, gain, black_level, exposure_time, wavelength,
# temperature, housing_temperature, fpn_correction,
# summary_only

op1 = report.info_op()
op1['summary_only'] = False
op1['camera_settings']['Gain'] = 0.1
op1['camera_settings']['Black level'] = 29.4
op1['camera_settings']['Bit depth'] = '12 bits'
op1['test_parameters']['Illumination'] = 'Variable with constant \
exposure time'
op1['test_parameters']['Irradiation steps'] = 50

# Add the operation point to the report
# we can add as many operation points as we want
# we pass the emva1288.Data1288 object to extract automatically all the results
# and graphics
myreport.add(op1, dat.data)

# Generate the latex files
myreport.latex()



import numpy as np


def qe(wavelength):
    """Simulate quantum efficiency for a specific wavelengths.

    Parameters
    ----------
    wavelength : float
                 The wavelength to compute the quantum efficency for.

    Returns
    -------
    float :
        The simulated quantum efficiency.
    """
    # For the time being we just simulate a simple gaussian
    s = 0.5
    u = 0.
    min_ = 350
    max_ = 800
    w = -1 + (wavelength) / (max_ - min_)
    qe = -.1 + np.exp((-(w - u) ** 2) /
                      (2 * s ** 2)) / (np.sqrt(np.pi * 2 * s ** 2))
    if qe < 0:
        return 0
    return qe


def nearest_value(value, array):
    """Returns the nearest value in vals.

    Parameters
    ----------
    value : float
            The value we want to get as near as possible.
    array : array_like
            The array containing the available values to get near the value.

    Returns
    -------
    The nearest element of `array` to `value`.
    """
    # http://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def get_irradiance(radiance, f):
    """Get The irradiance, in w/cm^2.

    Parameters
    ----------
    radiance : float
               The radiance (in W/sr/cm^2) to compute the irradiance from.
    f : float
        The f number of the setup.

    Returns
    -------
    float :
        The irradiance in W/cm^2
    """
    j = np.pi * radiance / (1 + ((2 * f) ** 2))
    return j


def get_photons(exposure, wavelength, radiance, pixel_area, f_number):
    """Get the number of photons hitting one pixel.

    Parameters
    ----------
    exposure : float
               The pixel exposure time (in ns) to the light.
    wavelength : float
                 The light wavelength hitting the pixel (in nm).
    radiance : float
               The radiance hitting the sensor (in W/sr/cm^2).
    pixel_area : float
                 The pixel area in um ^ 2
    f_number : float
               The f number of the setup.

    Returns
    -------
    float :
        The number of photons that hit the pixel.
    """
    h = 6.63e-34
    c = 3.00e8
    w = wavelength * 1e-9
    t = exposure * 1e-9
    a = pixel_area * 1e-12
    j = get_irradiance(radiance, f_number)
    return j * a * t * w / (h * c)


def get_radiance(exposure, wavelength, photons, pixel_area, f_number):
    """From the number of photons, get the radiance hitting a pixel.

    Parameters
    ----------
    exposure : float
               The pixel exposure time to the light (in ns).
    wavelength : float
                 The photons' wavelength (in nm).
    photons : float
              The number of photons that hit the pixel.
    pixel_area : float
                 The pixel area in um^2.
    f_number : float
               The f number of the setup.

    Returns
    -------
    float :
        The radiance that hit the pixel and gave the number of photons.
    """
    h = 6.63e-34
    c = 3.00e8
    w = wavelength * 1e-9
    t = exposure * 1e-9
    a = pixel_area * 1e-12

    #  p = j * a * t * w / (h * c)
    # j = np.pi * radiance / (1 + ((2 * f) ** 2))

    j = photons * h * c / (a * t * w)
    r = j * (1 + ((2 * f_number) ** 2)) / np.pi
    return r



import numpy as np
from emva1288.camera import routines


class Camera(object):
    """Camera simulator. It creates images according to the given parameters.
    """
    def __init__(self,
                 f_number=8,  # F-number of the light source/camera setup
                 pixel_area=25,  # um^2
                 bit_depth=8,  # Bit depth of the image [8, 10, 12, 14]
                 width=640,
                 height=480,

                 temperature=22,  # Sensor temperature in ^oC
                 temperature_ref=30,  # Reference temperature
                 temperature_doubling=8,  # Doubling temperature

                 wavelength=525,  # illumination wavelength
                 qe=None,  # Quantum efficiency for the given wavelength

                 exposure=1000000,  # Exposure time in ns
                 exposure_min=50000,  # Minimum exposure time in ns
                 exposure_max=500000000,  # Maximum exposure time in ns

                 K=0.1,  # Overall system gain
                 K_min=0.1,
                 K_max=17.,
                 K_steps=255,

                 blackoffset=0,
                 blackoffset_min=0,
                 blackoffset_max=None,
                 blackoffset_steps=255,

                 dark_current_ref=30,
                 dark_signal_0=0,
                 sigma2_dark_0=0,

                 dsnu=None,
                 prnu=None
                 ):
        """Camera simulator init method.

        Parameters
        ----------
        f_number : float, optional
                   The emva1288 f_number for the camera.
        pixel_area : float, optional
                     The area of one pixel (in um ^ 2)
        bit_depth : int, optiona
                    The number of bits allowed for one pixel value.
        width : int, optional
                The number of columns in the the image.
        height : int, optional
                The number of rows in the image.
        temperature : float, optional
                      The camera's sensor temperature in degrees Celsius.
        temperature_ref : float, optional
                          The reference temperature (at which the dark current
                          is equal to the reference dark current).
        temperature_doubling: float, optional
                              The doubling temperature (at which the dark
                              current is two times the reference dark
                              current).
        wavelength : float, optional
                     The light wavelength hitting the sensor (in nm).
        qe : float, optional
             Quantum efficiency (between 0 and 1). If None, a simulated
             quantum efficiency is choosen with the
             :func:`~emva1288.camera.routines.qe` function.
        exposure : float, optional
                   The camera's exposure time in ns.
        exposure_min : float, optional
                       The camera's minimal exposure time in ns.
        exposure_max : float, optional
                       The camera's maximal exposure time in ns.
        K : float, optional
            The overall system gain (in DN/e^-).
        K_min : float, optional
                The overall minimal system gain (in DN/e^-).
        K_max : float, optional
                The overall maximal system gain (in DN/e^-).
        K_steps : int, optional
                  The number of available intermediate overall system gains
                  between K_min and K_max.
        blackoffset : float, optional
                      The dark signal offset for each pixel (in DN).
        blackoffset_min: float, optional
                         The minimal dark signal offset for each pixel (in DN).
        blackoffset_max : float, optional
                          The maximal dark signal offset for each pixel
                          (in DN).
        blackoffset_steps : int, optional
                            The number of available blackoffsets between the
                            mimimal and maximal blackoffsets.
        dark_current_ref : float, optional
                           The reference dark current used for computing the
                           total dark current.
        dark_signal_0 : float, optional
                        The offset for the computation of the mean number
                        of electrons in the dark for one pixel (in DN/s).
        sigma2_dark_0 : float, optional
                        The offset for the computation of the dark signal
                        (the dark signal standart deviation).
        dsnu : np.array, optional
               DSNU image in DN, array with the same shape of the image
               that is added to every image
        prnu : np.array, optional
               PRNU image in percentages (1 = 100%), array with the same shape
               of the image. Every image is multiplied by it
        """

        self._pixel_area = pixel_area
        self._bit_depth = bit_depth
        self._img_max = 2 ** int(bit_depth) - 1
        self._width = width
        self._height = height

        self._temperature_ref = temperature_ref
        self._temperature_doubling = temperature_doubling

        self._qe = qe
        # When no specific qe is provided we simulate one
        if qe is None:
            self._qe = routines.qe(wavelength)

        self._dark_current_ref = dark_current_ref
        self._dark_signal_0 = dark_signal_0
        self._sigma2_dark_0 = sigma2_dark_0

        self._exposure = exposure
        self._exposure_min = exposure_min
        self._exposure_max = exposure_max

        self.__Ks = np.linspace(K_min, K_max, num=K_steps)
        self._K = None
        self.K = K

        # A good gestimate for maximum blackoffset is 1/16th of the full range
        if not blackoffset_max:
            blackoffset_max = self.img_max // 16
        self.__blackoffsets = np.linspace(blackoffset_min, blackoffset_max,
                                          num=blackoffset_steps)
        self._blackoffset = None
        self.blackoffset = blackoffset

        self._dsnu = dsnu
        self._prnu = prnu
        if dsnu is None:
            self._dsnu = np.zeros((self.height, self.width))
        if prnu is None:
            self._prnu = np.ones((self.height, self.width))
        self.environment = {'temperature': temperature,
                            'wavelength': wavelength,
                            'f_number': f_number}

    @property
    def bit_depth(self):
        """The number of bits allowed for a gray value for one pixel."""
        return self._bit_depth

    @property
    def pixel_area(self):
        """The area of one pixel (in um ^ 2)."""
        return self._pixel_area

    @property
    def img_max(self):
        """The maximal value for one pixel (in DN)."""
        return self._img_max

    @property
    def width(self):
        """The number of columns"""
        return self._width

    @property
    def height(self):
        """The number of rows"""
        return self._height

    @property
    def exposure(self):
        """The camera's exposure time (in ns)."""
        return self._exposure

    @exposure.setter
    def exposure(self, value):
        self._exposure = value

    @property
    def exposure_min(self):
        """The camera's minimal exposure time (in ns)."""
        return self._exposure_min

    @property
    def exposure_max(self):
        """The camera's maximal exposure time (in ns)."""
        return self._exposure_max

    @property
    def K(self):
        """The overall system gain (in DN/e^-).

        :Setter: The setter uses the
                 :func:`~emva1288.camera.routines.nearest_value` function
                 to set the system gain to the nearest value given to the
                 setter.
                 This is because not all system gains are possible but rather a
                 linear sample between the minimal and maximal value.
        """
        return self._K

    @K.setter
    def K(self, value):
        self._K = routines.nearest_value(value, self.__Ks)

    @property
    def Ks(self):
        """The array of all the available system gains (in DN/e^-)."""
        return self.__Ks

    @property
    def blackoffset(self):
        """The system dark signal offset (in DN).

        :Setter: The setter uses the
                 :func:`~emva1288.camera.routines.nearest_value` function
                 to set the black signal offset to the nearest value given
                 to the setter. This is because not all black offsets are
                 possible but rather a linear sample between the minimal
                 and maximal value.
        """
        return self._blackoffset

    @blackoffset.setter
    def blackoffset(self, value):
        self._blackoffset = routines.nearest_value(value,
                                                   self.__blackoffsets)

    @property
    def blackoffsets(self):
        """The array of all blackoffsets (in DN)."""
        return self.__blackoffsets

    def grab(self, radiance, temperature=None, wavelength=None, f_number=None):
        """
        Create an image based on the mean and standard deviation from the
        EMVA1288 parameters.

        The image is generated using a normal distribution for each pixel.

        Parameters
        ----------
        radiance : float
                   The sensor's illumination in W/cm^2/sr. This is the only
                   mandatory argument because it is frequently changed during
                   an test.
        temperature : float, optional
                      The camera's temperature in degrees Celsius.
                      If None, the environment's
                      temperature will be taken.
        wavelength : float, optional
                     The illumination wavelength in nanometers.
                     If None, the environment's wavelength will be taken.
        f_number : float, optional
                   The optical setup f_number.
                   If None, the environment's f_number will be taken.
        """
        clipping_point = int(self.img_max)
        u_y = self._u_y(radiance,
                        temperature=temperature,
                        wavelength=wavelength,
                        f_number=f_number)
        s2_y = np.sqrt(self._s2_y(radiance,
                                  temperature=temperature,
                                  wavelength=wavelength,
                                  f_number=f_number))
        img = np.random.normal(loc=u_y, scale=s2_y,
                               size=(self.height, self.width))

        # not the best but hope it works as approach for prnu dsnu
        img *= self._prnu
        img += self._dsnu

        img += self.blackoffset
        np.rint(img, img)
        np.clip(img, 0, clipping_point, img)
        return img.astype(np.uint64)

    def _u_y(self, radiance, temperature=None, wavelength=None, f_number=None):
        """
        Mean digital value (in DN) of the image.
        """
        uy = self.K * (self._u_d(temperature=temperature) +
                       self._u_e(radiance,
                                 wavelength=wavelength,
                                 f_number=f_number))
        return uy

    def _u_e(self, radiance, wavelength=None, f_number=None):
        """
        Mean number of electrons per pixel during exposure time.
        """
        u_e = self._qe * self.get_photons(radiance, wavelength=wavelength,
                                          f_number=f_number)
        return u_e

    def _s2_e(self, radiance, wavelength=None, f_number=None):
        """
        Variance of the number of electrons.

        Same as u_e because the number of electrons is supposed to
        be distributed by a Poisson distribution where the mean
        equals the variance.
        """
        return self._u_e(radiance, wavelength=wavelength, f_number=f_number)

    def _u_d(self, temperature=None):
        """
        Mean number of electrons without light.
        """
        u_d = ((self._u_i(temperature=temperature) *
                self.exposure / (10 ** 9)) +
               self._dark_signal_0)
        return u_d

    def _s2_q(self):
        """
        Variance of the quantization noise.
        """
        return 1.0 / 12.0

    def _s2_y(self, radiance, temperature=None, wavelength=None,
              f_number=None):
        """
        Variance of the digital signal (= temporal noise).
        """
        s2_y = ((self.K ** 2) * (self._s2_d(temperature=temperature) +
                self._s2_e(radiance,
                           wavelength=wavelength,
                           f_number=f_number)) +
                self._s2_q())
        return s2_y

    def _u_i(self, temperature=None):
        """
        Dark current (in DN/s).
        """
        if temperature is None:
            temperature = self.environment['temperature']
        u_i = 1. * self._dark_current_ref * 2 ** (
            (temperature - self._temperature_ref) / self._temperature_doubling)
        return u_i

    def _s2_d(self, temperature=None):
        """
        Variance of the dark signal =  Dark temporal noise.
        """
        s2_d = self._sigma2_dark_0 + (self._u_i(temperature=temperature) *
                                      self.exposure / (10 ** 9))
        return s2_d

    def get_radiance_for(self, mean=None, exposure=None):
        """Radiance to achieve saturation.

        Calls the :func:`~emva1288.camera.routines.get_radiance` function
        to get the radiance for saturation.

        Parameters
        ----------
        mean : float, optional
               The saturation value of the camera. If None, this value
               is set to the :attr:`img_max` attribute.
        exposure : float, optional
                   The camera's exposure time at which the radiance
                   for saturation value will be computed. If None, the
                   exposure time taken will be the camera's actual exposure
                   time.

        Returns
        -------
        float :
            The radiance at which, for the given saturation value and
            the given exposure time, the camera saturates.
        """
        if not mean:
            mean = self.img_max
        if not exposure:
            exposure = self.exposure
        ud = self._u_d()
        ue = (mean / self.K) - ud
        up = ue / self._qe
        radiance = routines.get_radiance(exposure,
                                         self.environment['wavelength'],
                                         up,
                                         self.pixel_area,
                                         self.environment['f_number'])
        return radiance

    def get_photons(self, radiance, exposure=None,
                    wavelength=None, f_number=None):
        """Computes the number of photons received by one pixel.

        Uses the :func:`~emva1288.camera.routines.get_photons` function
        to compute this number.

        Parameters
        ----------
        radiance : float
                   The radiance exposed to the camera (in Wsr^-1cm^-2).
        exposure : float, optional
                   The pixel's exposure time in ns.

        Returns
        -------
        float :
            The number of photons received by one pixel.
        """
        if exposure is None:
            exposure = self.exposure
        if f_number is None:
            f_number = self.environment['f_number']
        if wavelength is None:
            wavelength = self.environment['wavelength']
        return routines.get_photons(exposure,
                                    wavelength,
                                    radiance,
                                    self.pixel_area,
                                    f_number)



"""
Calling one by one the classes that compose the reference implementation
process a descriptor file, print the results and plot the graphics
"""

from emva1288 import process
from emva1288.camera.dataset_generator import DatasetGenerator

# dir_ = '/home/work/1288/datasets/'
# fname = 'EMVA1288_ReferenceSet_003_Simulation_12Bit/EMVA1288_Data.txt'

dataset_generator = DatasetGenerator(width=100, height=50, bit_depth=8,
                                     dark_current_ref=30)

# Parse the descriptor file
# parser = process.ParseEmvaDescriptorFile(os.path.join(dir_, fname))
parser = process.ParseEmvaDescriptorFile(dataset_generator.descriptor_path)

# Load images
imgs = process.LoadImageData(parser.images)

# Extract data from images
dat = process.Data1288(imgs.data)

# Compute the results
res = process.Results1288(dat.data)
res.print_results()

# Plot the graphics
plot = process.Plotting1288(res)
plot.plot()



import numpy as np
from emva1288.camera import Camera

c = Camera()

radiances = np.linspace(0, c.get_radiance_for(mean=200), num=20)
for radiance in radiances:
    img = c.grab(radiance)
    print(img.mean(), img.std())



import unittest
from emva1288.process.parser import ParseEmvaDescriptorFile
from emva1288.camera.dataset_generator import DatasetGenerator


class TestParser(unittest.TestCase):
    # attrbutes for dataset generator
    _bit_depth = 8
    _version = '3.0'
    _height = 50
    _width = 100
    _L = 50
    _steps = 10

    # Don't test error raised because those errors appears when descriptor file
    # is not well formatted and not because of a code failure.

    def setUp(self):
        # create data descriptor file for parser
        self.d_generator = DatasetGenerator(bit_depth=self._bit_depth,
                                            height=self._height,
                                            width=self._width,
                                            L=self._L,
                                            version=self._version,
                                            steps=self._steps)

    def tearDown(self):
        # delete generator to delete all the generated files
        del self.d_generator

    def test_good_descriptorfile(self):
        # test that the parser actually parses the file with the generated file
        descriptor_file = self.d_generator.descriptor_path
        parser = ParseEmvaDescriptorFile(descriptor_file)
        # data manually taken from the file:
        bits = self._bit_depth
        height = self._height
        width = self._width
        times = self.d_generator.points['temporal'].keys()
        first_exp_time = list(times)[0]
        first_rad = self.d_generator.points['temporal'][first_exp_time][0]
        first_pcount = round(self.d_generator.cam.get_photons(first_rad,
                                                              first_exp_time),
                             3)

        # check data have correctly been parsed
        self.assertEqual(parser.version, self._version)
        self.assertEqual(parser.format['bits'], bits)
        self.assertEqual(parser.format['height'], height)
        self.assertEqual(parser.format['width'], width)
        # for this expTime and pcount, there is only 2 images thus temporal
        im = parser.images['temporal'][first_exp_time][first_pcount]
        # The length of this dict should be 2
        self.assertEqual(len(im), 2)

        # For spatial data
        points = self.d_generator.points['spatial']
        spatial_texp = list(points.keys())[0]
        spatial_rad = list(points.values())[0][0]
        # round here because pcount are rounded in descriptor file
        spatial_pcount = round(self.d_generator.cam.get_photons(spatial_rad),
                               3)
        im_spatial = parser.images['spatial'][spatial_texp][spatial_pcount]
        # the length of this dict should be greater than 2
        self.assertGreater(len(im_spatial), 2)

        # For dark images, pcount should be 0
        # for this time, there is a dark image
        # dark images are normal images with 0.0 photon count
        keys = parser.images['temporal'][first_exp_time].keys()
        self.assertTrue(0.0 in keys)



from emva1288.report.report import info_setup
from emva1288.report.report import info_basic
from emva1288.report.report import info_marketing
from emva1288.report.report import info_op
from emva1288.report.report import Report1288



import nose
import os


default_test_modules = ['emva1288.unittests.test_coding_standards',
                        'emva1288.unittests.test_camera',
                        'emva1288.unittests.test_parser.TestParser',
                        'emva1288.unittests.test_loader.TestLoader',
                        'emva1288.unittests.test_data.TestData',
                        'emva1288.unittests.test_results.TestResults',
                        'emva1288.unittests.test_report.TestReportGenerator']


def run():
    os.environ["NOSE_WITH_COVERAGE"] = "1"
    os.environ['NOSE_COVER_PACKAGE'] = 'emva1288'
    os.environ["NOSE_COVER_HTML"] = "1"
    os.environ["NOSE_COVER_ERASE"] = "1"
    nose.main(defaultTest=default_test_modules)

if __name__ == '__main__':
    run()



import unittest
import emva1288.camera as cam
import numpy as np


class CameraTestCase(unittest.TestCase):
    def setUp(self):
        self.cam = cam.Camera()

    def tearDown(self):
        del self.cam

    def test_img(self):
        img = self.cam.grab(0)
        self.assertEqual((self.cam.height, self.cam.width), np.shape(img))

    def test_radiance(self):
        img1 = self.cam.grab(0)
        img2 = self.cam.grab(self.cam.get_radiance_for(mean=250))
        self.assertLess(img1.mean(), img2.mean())



# -*- coding: utf-8 -*-
# Copyright (c) 2014 The EMVA1288 Authors. All rights reserved.
# Use of this source code is governed by a GNU GENERAL PUBLIC LICENSE that can
# be found in the LICENSE file.

"""EMVA 1288 descriptor parser
This class takes an EMVA1288 descriptor file and loads its content into a
python dictionary.

An EMVA1288 descriptor file is a file that contains the description of an
EMVA1288 test including exposure times, photon count and corresponding images
"""

import numpy as np
import os
import pprint
import logging


class ParseEmvaDescriptorFile(object):
    """Take an image descriptor file and transform it into
    an usable directory
    """

    def __init__(self, filename, path=None, loglevel=logging.INFO):
        """Parser init method.

        Uses a :class:`python:logging.Logger` object to print infos of the
        parcing process. This method :meth:`loads <_load_file>` the file
        and :meth:`gets <_fill_info>` the information from it.

        Parameters
        ----------
        filename : str
                   The descriptor file's name or the complete path to it.
        path : str, optional
               The path to the descriptor file.
        loglevel : int, optional
                   The logger level.
        """
        # The items are in the form of
        # exposure:{photons:[fname1, fname2,...]}, photons....}
        # for dark, the number of photons is 0.0

        # If no path is given, the filename path will be used to fill
        # the images dict
        self._path = path

        self.format = {}  # bits, witdth, height
        self.version = None
        self.images = {'temporal': {},
                       'spatial': {}}

        logging.basicConfig()
        self.log = logging.getLogger('Parser')
        self.log.setLevel(loglevel)

        self._load_file(filename)
        self._fill_info()
        self.log.debug(pprint.pformat(self.images))

    def _get_images_filenames(self):
        """
        From the current line in self._lines array
        get all the consecutive "i filename"
        if less than 2 consecutive, raise an error
        """
        fnames = []
        while self._lines:
            line = self._lines.pop()
            l = self._split_line(line)
            if l[0] != 'i':
                # Ups, to the end of images, reappend last line that is not an
                # image line
                self._lines.append(line)
                break
            if len(l) != 2:  # pragma: no cover
                raise SyntaxError('Wrong format: "%s" should be "i filename"'
                                  % line)
                break
            # append image path to fnames
            npath = os.path.normpath(l[1])
            path = os.path.join(self._path, *npath.split('\\'))
            fnames.append(path)

        if len(fnames) < 2:  # pragma: no cover
            raise SyntaxError('Each image series, has to '
                              'have at least two images')

        return fnames

    def _get_kind(self, fnames):
        """
        Guess what kind of data based on the number of images
        Temporal = 2 images for each measurement point
        Spatial = >2 images for each measurement point
        """
        L = len(fnames)
        if L == 2:
            kind = 'temporal'
        else:
            kind = 'spatial'

        return kind

    def _add_pcount(self, exposure, photons, fnames):
        """Add images to a given exposure/phton

        For a given exposure and photon count
        add the appropiate image filenames to the self.images dict
        """
        # is it temporal or spatial data
        kind = self._get_kind(fnames)
        # create the exposure time dictionary for this exposure time
        # if it is not already existing
        self.images[kind].setdefault(exposure, {})

        # if this dict for this exposure time and this
        # photons count already existed, raise an error in order to not
        # overwrite existing data.
        if photons in self.images[kind][exposure]:  # pragma: no cover
            raise SyntaxError('Only one set of images exp %.3f photons %.3f'
                              % (exposure, photons))

        # append the images path to a dict whose key is the photons count
        # inside the exposure time dict
        self.images[kind][exposure][photons] = fnames

    def _fill_info(self):
        """
        Iterate through all the lines in the descriptor file
        and parse them by their first character. Fill self.images
        """

        # Start at the end of the file
        self._lines.reverse()
        while self._lines:
            # pop it such that other methods know the current processed line
            line = self._lines.pop()
            # check line is good format and split it
            l = self._split_line(line)

            # descriptor file supposed format ##
            # n bits width height
            # b exposureTime(ns) numberPhotons  (bright image)
            # i relativePathToTheImage
            # d exposureTime(ns)                (dark image)

            if l[0] == 'v':
                # if line starts with 'v', this is the version
                self.version = l[1]
                self.log.info('Version ' + l[1])
                continue

            if l[0] == 'n':
                # for lines that starts with n, there is always 4 elements
                # n + bits + width + height
                # There should be only one of this line in the file
                if len(l) != 4:  # pragma: no cover
                    raise SyntaxError('Wrong format: "%s" should be "n bits '
                                      'width height"' % line)

                if self.format:  # pragma: no cover
                    # if it is the second line found
                    # of this type raise error
                    raise SyntaxError('Only one "n bits width height" is '
                                      'allowed per file')

                self.format['bits'] = int(l[1])
                self.format['width'] = int(l[2])
                self.format['height'] = int(l[3])
                continue

            if l[0] == 'b':
                # For lines that starts with b. there is always 3 elements
                # b + exposureTime + numberPhotons (bright images)
                if len(l) != 3:  # pragma: no cover
                    raise SyntaxError('Wrong format: "%s" should be "b '
                                      'exposure photons"' % line)

                # Replace floating point representation if wrong format.
                exposure = np.float(l[1].replace(',', '.'))
                photons = np.float(l[2].replace(',', '.'))
                # For this settings, get all the corresponding images
                fnames = self._get_images_filenames()
                # Add the images path to the images[kind][exposure][photons]
                # dictionary where kind = temporal or spatial
                self._add_pcount(exposure, photons, fnames)

                continue

            if l[0] == 'd':
                # For lines that starts with d, there is always 2 elements
                # d + exposureTime (dark images)
                if len(l) != 2:  # pragma: no cover
                    raise SyntaxError('Wrong format: "%s" should be "d '
                                      'exposure"' % line)

                # replace floating point representation if wring format
                exposure = np.float(l[1].replace(',', '.'))
                # For this exposure, get all the corresponding images
                fnames = self._get_images_filenames()
                # Add the images path to the images dict.
                self._add_pcount(exposure, np.float(0.0), fnames)

                continue

            # If line is of an unknown format, warn user.
            self.log.warning('Unknown command ' + line)  # pragma: no cover

    def _split_line(self, line):
        """
        For every line of descriptorfile
        check that it has at least two arguments
        split by white spaces and strip white spaces from elements
        """
        l = [x.strip() for x in line.split()]
        if (not l) or (len(l) < 2):  # pragma: no cover
            raise SyntaxError('Wrong format line: %s' % line)
        return l

    def _load_file(self, filename):
        """
        Load a file, split by lines removing the comments (starts with #)
        """
        self.log.info('Opening ' + filename)
        f = open(filename, 'r')

        # To add location when opening images
        # If no path was passed as kwarg, set it to the filename path
        if self._path is None:
            self._path = os.path.dirname(filename)

        # get the lines and strip them if they are not comments
        try:
            self._lines = [x.strip() for x in f.readlines() if x.strip() and
                           not x.strip().startswith('#')]
        except UnicodeDecodeError:  # pragma: no cover
            # If there is an unknown character in the file, speak it!
            raise UnicodeDecodeError("File: '%s', has non-utf8 characters."
                                     "Find them and kill them!" % filename)



import unittest
from emva1288.camera.dataset_generator import DatasetGenerator
from emva1288.process.parser import ParseEmvaDescriptorFile
from emva1288.process.loader import LoadImageData


class TestLoader(unittest.TestCase):
    _height = 50
    _width = 100
    _bit_depth = 8
    _L = 50
    _steps = 10

    def _init(self):
        # Create dataset to load
        dataset = DatasetGenerator(height=self._height,
                                   width=self._width,
                                   bit_depth=self._bit_depth,
                                   L=self._L,
                                   steps=self._steps)
        descriptor_path = dataset.descriptor_path
        # create the parser
        parser = ParseEmvaDescriptorFile(descriptor_path)
        # create loader
        loader = LoadImageData(parser.images)
        return dataset, parser, loader

    def test_loader(self):
        d, p, l = self._init()
        self.dataset = d
        self.parser = p
        self.loader = l
        # Test that checks if loader actually loads data from images given
        # by the parser

        # test that the data attribute contains the good infos
        data = self.loader.data
        self.assertEqual(data['height'], self._height)
        self.assertEqual(data['width'], self._width)
        first_exp_time = self.dataset.cam.exposure_min
        # temporal data should contain 2 datasets (one bright one dark)
        temporal_data = data['temporal'][first_exp_time]
        self.assertEqual(len(temporal_data), 2)
        self.assertTrue(0.0 in temporal_data.keys())
        # there should be steps data sets for temporal
        self.assertEqual(len(self.loader.data['temporal']), self._steps)

        spatial_texp = list(self.dataset.points['spatial'].keys())[0]

        # spatial data should contain 2 sets (one dark and one bright)
        spatial_data = data['spatial'][spatial_texp]
        self.assertEqual(len(spatial_data), 2)
        self.assertTrue(0.0 in spatial_data.keys())
        # data should be made of L images
        self.assertEqual(spatial_data[0.0]['L'], self._L)

        # check data type and format
        for typ in ('sum', 'pvar'):
            # data is sum and pvar
            self.assertTrue(typ in spatial_data[0.0].keys())
            self.assertTrue(typ in temporal_data[0.0].keys())
            # spatial data is sum images and pvar
            self.assertEqual(spatial_data[0.0][typ].shape, (self._height,
                                                            self._width))

        del self.parser
        del self.dataset
        del self.loader

    def test_loader_errors(self):
        # check that images with no dark images raise ValueError
        with self.assertRaises(ValueError):
            images = {'temporal': {0: {0.1: ""}},
                      'spatial': {0: {0.1: ""}}}
            l = LoadImageData(images)
        # check that one image for temporal instead of 2 raise valueerror
        with self.assertRaises(ValueError):
            images = {'temporal': {0: {0.0: ""}},
                      'spatial': {0: {0.0: ""}}}
            l = LoadImageData(images)

        # Check that an image that does not exist raise an IOError
        with self.assertRaises(IOError):
            images = {'temporal': {0: {0.0: ["."], 0.1: ["."]}},
                      'spatial': {0: {0.0: ["."], 0.1: ["."]}}}
            l = LoadImageData(images)



import pep8
import os
import emva1288
from nose.tools import assert_equal

PEP8_ADDITIONAL_IGNORE = []
EXCLUDE_FILES = []


def test_pep8_conformance():

    dirs = []
    dirname = os.path.dirname(emva1288.__file__)
    dirs.append(dirname)
    examplesdir = os.path.join(dirname, '..', 'examples')
    examplesdir = os.path.abspath(examplesdir)
    dirs.append(examplesdir)

    pep8style = pep8.StyleGuide()

    # Extend the number of PEP8 guidelines which are not checked.
    pep8style.options.ignore = (pep8style.options.ignore +
                                tuple(PEP8_ADDITIONAL_IGNORE))
    pep8style.options.exclude.extend(EXCLUDE_FILES)

    result = pep8style.check_files(dirs)
    msg = "Found code syntax errors (and warnings)."
    assert_equal(result.total_errors, 0, msg)

if __name__ == '__main__':
    import nose
    nose.runmodule()
